<h3>Download</h3>
{% if version.endswith('(hg)') %}
<p>This documentation is for version <b>{{ version }}</b>, which is
  not released yet.</p>
<p>You can use it from the
  <a href="http://code.google.com/p/pytox/source/checkout">Mercurial repo</a> or look for
  released versions in the <a href="http://pypi.python.org/pypi/tox">Python
    Package Index</a>.</p>
{% else %}
<p>Current: <b>{{ version }}</b>
[<a href="{{ pathto('changelog') }}">Changes</a>]</p>
<p>
<a href="http://pypi.python.org/pypi/tox">tox on PyPI</a>
</p>
<pre>pip install tox</pre>
{% endif %}

<h3>Questions? Suggestions?</h3>

<p>Checkout <a href="{{ pathto('support') }}">support channels</a>
</p>
